
import numpy
import subprocess
import tempfile
import os

numpy.set_printoptions(linewidth=150, precision=8, suppress=False)

def read_problem_file(filename):
    f = open(filename, 'r')

    line = f.readline()
    line = line.strip()
    num_constraints = int(line)

    line = f.readline()
    line = line.strip()
    num_blocks = int(line)

    line = f.readline()
    parts = line.split()
    block_sizes = []
    for part in parts:
        block_sizes.append(int(part))

    blocks = []
    for size in block_sizes:
        if size < 0:
            block = numpy.zeros([1, -size, num_constraints + 1], float)
        else:
            block = numpy.zeros([size, size, num_constraints + 1], float)
        blocks.append(block)

    line = f.readline()
    parts = line.split()
    constants = numpy.zeros([num_constraints], float)
    for i in xrange(num_constraints):
        constants[i] = float(parts[i])

    while True:
        line = f.readline()
        if len(line) == 0:
            break
        parts = line.split()
        if len(parts) == 0:
            continue
        constraint = int(parts[0])
        block_num = int(parts[1])
        row = int(parts[2])
        column = int(parts[3])
        value = float(parts[4])

        if block_sizes[block_num - 1] < 0:
            blocks[block_num - 1][0, column - 1, constraint] += value
        else:
            blocks[block_num - 1][row - 1, column - 1, constraint] += value
            if row != column:
                blocks[block_num - 1][column - 1, row - 1, constraint] += value


    problem = dict()
    problem['num_constraints'] = num_constraints
    problem['num_blocks'] = num_blocks
    problem['block_sizes'] = block_sizes
    problem['blocks'] = blocks
    problem['constants'] = constants
    f.close()
    return problem


def read_problem_file_as_int(filename):
    f = open(filename, 'r')

    line = f.readline()
    line = line.strip()
    num_constraints = int(line)

    line = f.readline()
    line = line.strip()
    num_blocks = int(line)

    line = f.readline()
    parts = line.split()
    block_sizes = []
    for part in parts:
        block_sizes.append(int(part))

    blocks = []
    for size in block_sizes:
        if size < 0:
            block = numpy.zeros([1, -size, num_constraints + 1], int)
        else:
            block = numpy.zeros([size, size, num_constraints + 1], int)
        blocks.append(block)

    line = f.readline()
    parts = line.split()
    constants = numpy.zeros([num_constraints], int)
    for i in xrange(num_constraints):
        constants[i] = int(parts[i])

    while True:
        line = f.readline()
        if len(line) == 0:
            break
        parts = line.split()
        if len(parts) == 0:
            continue
        constraint = int(parts[0])
        block_num = int(parts[1])
        row = int(parts[2])
        column = int(parts[3])
        value = int(parts[4])

        if block_sizes[block_num - 1] < 0:
            blocks[block_num - 1][0, column - 1, constraint] += value
        else:
            blocks[block_num - 1][row - 1, column - 1, constraint] += value
            if row != column:
                blocks[block_num - 1][column - 1, row - 1, constraint] += value


    problem = dict()
    problem['num_constraints'] = num_constraints
    problem['num_blocks'] = num_blocks
    problem['block_sizes'] = block_sizes
    problem['blocks'] = blocks
    problem['constants'] = constants
    f.close()
    return problem


def write_problem_file(filename, problem):
    f = open(filename, 'w')

    num_constraints = problem['num_constraints']
    num_blocks = problem['num_blocks']
    block_sizes = problem['block_sizes']
    blocks = problem['blocks']
    constants = problem['constants']

    f.write("%d\n" % num_constraints)
    f.write("%d\n" % num_blocks)
    for size in block_sizes:
        f.write("%d " % size)
    f.write("\n")

    for constant in constants:
        f.write("%.13g " % constant)
    f.write("\n")

    def print_row():
        if value != 0:
            f.write("%d %d %d %d %.13g\n" % (constraint, block_num + 1, row + 1, column + 1, value))

    for constraint in xrange(num_constraints + 1):
        for block_num in xrange(num_blocks):
            block_size = block_sizes[block_num]
            if block_size < 0:
                for column in xrange(-block_size):
                    row = column
                    value = blocks[block_num][0, column, constraint]
                    print_row()
            else:
                for row in xrange(block_size):
                    for column in xrange(row, block_size):
                        value = blocks[block_num][row, column, constraint]
                        print_row()

    f.close()



def read_solution_file(filename, problem):
    f = open(filename, 'r')

    num_constraints = problem['num_constraints']
    num_blocks = problem['num_blocks']
    block_sizes = problem['block_sizes']

    Y = numpy.zeros([num_constraints], float)
    line = f.readline()
    parts = line.split()
    for i in xrange(num_constraints):
        Y[i] = float(parts[i])

    Primal = []
    Dual = []

    for block_num in xrange(num_blocks):
        if block_sizes[block_num] < 0:
            Primal.append(numpy.zeros([1, -block_sizes[block_num]], float))
            Dual.append(numpy.zeros([1, -block_sizes[block_num]], float))
        else:
            Primal.append(numpy.zeros([block_sizes[block_num], block_sizes[block_num]], float))
            Dual.append(numpy.zeros([block_sizes[block_num], block_sizes[block_num]], float))

    while True:
        line = f.readline()
        if len(line) == 0:
            break
        parts = line.split()
        if len(parts) == 0:
            continue
        constraint = int(parts[0])
        block_num = int(parts[1])
        row = int(parts[2])
        column = int(parts[3])
        value = float(parts[4])

        if constraint == 2:
            block = Primal[block_num - 1]
        else:
            block = Dual[block_num - 1]

        if block_sizes[block_num - 1] < 0:
            block[0, column - 1] += value
        else:
            block[row - 1, column - 1] += value
            if row != column:
                block[column - 1, row - 1] += value


    solution = dict()
    solution['Y'] = Y
    solution['Primal'] = Primal
    solution['Dual'] = Dual
    return solution

def extract_problem_matrix(problem, constants = True, objective = True, first_block = True, surpluses = True):
    num_constraints = problem['num_constraints']
    num_blocks = problem['num_blocks']
    block_sizes = problem['block_sizes']

    total_variables = 0
    start = []
    for block_num in xrange(num_blocks):
        if not first_block and block_num == 0:
            start.append(0)
            continue

        start.append(total_variables)
        if not surpluses and block_num == 0:
            size = 1
        else:
            size = block_sizes[block_num]

        if size > 0:
            total_variables += size * (size + 1) / 2
        else:
            total_variables -= size

    blocks = problem['blocks']
    problem_constants = problem['constants']

    if objective:
        rows = num_constraints + 1
    else:
        rows = num_constraints

    if constants:
        columns = total_variables + 1
    else:
        columns = total_variables
    M = numpy.zeros([rows, columns])

    for constraint in xrange(num_constraints + 1):
        if objective:
            row = constraint
        else:
            if constraint == 0:
                continue
            row = constraint - 1

        if constants:
            if constraint == 0:
                M[row, total_variables] = 0
            else:
                M[row, total_variables] = problem_constants[constraint - 1]


        for block_num in xrange(num_blocks):
            if not first_block and block_num == 0:
                continue
            if not surpluses and block_num == 0:
                size = 1
            else:
                size = block_sizes[block_num]

            if size > 0:
                pos = 0
                for i in xrange(size):
                    for j in xrange(i, size):
                        value = blocks[block_num][i, j, constraint]
                        if i == j:
                            M[row, start[block_num] + pos] = value
                        else:
                            M[row, start[block_num] + pos] = 2 * value
                        pos += 1
            else:
                for j in xrange(-size):
                    value = blocks[block_num][0, j, constraint]
                    M[row, start[block_num] + j] = value

    return M


def extract_solution_vector(primal, first_block = True, surpluses = True):
    v = []
    for block_num in xrange(len(primal)):
        if not first_block and block_num == 0:
            continue
        block = primal[block_num]
        if not surpluses and block_num == 0:
            v.append(block[0, 0])
        for row in xrange(block.shape[0]):
            for column in xrange(row, block.shape[1]):
                v.append(block[row, column])

    return numpy.array(v, dtype=float)

def compact_solution_vector(problem, v, first_block = True, surpluses = True):
    primal = []
    pos = 0
    for block_num in xrange(problem['num_blocks']):
        size = problem['block_sizes'][block_num]
        if size < 0:
            block = numpy.zeros([1, -size], float)
        else:
            block = numpy.zeros([size, size], float)

        if not first_block and block_num == 0:
            primal.append(block)
            continue

        if not surpluses and block_num == 0:
            block[0,0] = v[0]
            pos += 1
            primal.append(block)
            continue

        for row in xrange(block.shape[0]):
            for column in xrange(row, block.shape[1]):
                block[row, column] = v[pos]
                if size > 0:
                    block[column,row] = v[pos]
                pos += 1
        primal.append(block)

    return primal

def find_variable_index(problem, block_num, row, column, first_block = True, surpluses = True):
    index = 1
    for i in xrange(block_num - 1):
        if i == 0:
            if not first_block: continue
            if not surpluses:
                index += 1
                continue

        size = problem['block_sizes'][i]
        if size < 0:
            index += -size
        else:
            index += size * (size + 1) /2

    if column < row:
        row, column = column, row

    size = problem['block_sizes'][block_num - 1]
    column = column - row
    while row > 1:
        index += size
        size -= 1
        row -= 1

    index += column
    return index

def find_variable_position(problem, index, first_block = True, surpluses = True):
    index -= 1
    for block_num in xrange(problem['num_blocks']):
        if block_num == 0:
            if not first_block: continue
            if not surpluses:
                if index > 0:
                    index -= 1
                    continue
                else:
                    break

        size = problem['block_sizes'][block_num]
        if size < 0:
            size = -size
        else:
            size = size * (size + 1) /2

        if index < size: break
        index -= size

    size = problem['block_sizes'][block_num]
    if size > 0:
        row = 1
        while index > 0:
            if index < size: break
            index -= size
            size -= 1
            row += 1

        column = row + index
    else:
        row = 1
        column = row + index
    return block_num + 1, row, column

def solve_problem(problem):
    problemfile, problemfilename = tempfile.mkstemp()
    os.close(problemfile)

    solutionfile, solutionfilename = tempfile.mkstemp()
    os.close(solutionfile)

    write_problem_file(problemfilename, problem)
    returncode = 0
    output = ""
    solution = None

    try:
        output = subprocess.check_output(["csdp", problemfilename, solutionfilename],
                                         stderr=subprocess.STDOUT)
        solution = read_solution_file(solutionfilename, problem)
    except subprocess.CalledProcessError as excpt:
        returncode = excpt.returncode
        output = excpt.output
        solution = read_solution_file(solutionfilename, problem)
    finally:
        os.remove(problemfilename)
        os.remove(solutionfilename)

    return returncode, output, solution


def solution_eigenvalues(primal, first_block = True, surpluses = True):
    v = []
    for block_num in xrange(len(primal)):
        if not first_block and block_num == 0:
            continue
        block = primal[block_num]
        if not surpluses and block_num == 0:
            v.append(block[0, 0])
        if block.shape[0] == 1:
            for column in xrange(block.shape[1]):
                v.append(block[0, column])
        else:
            eig = numpy.linalg.eigh(block)
            for row in xrange(len(eig[0])):
                v.append(eig[0][row])

    return numpy.array(v, dtype=float)

def solution_slackness(problem, primal, first_block = True, surpluses = True):
    num_constraints = problem['num_constraints']
    num_blocks = problem['num_blocks']
    blocks = problem['blocks']
    constants = problem['constants']

    v = []
    for constraint in xrange(1, num_constraints + 1):
        total = 0.0
        for block_num in xrange(num_blocks):
            if not first_block and block_num == 0:
                continue
            block = blocks[block_num]
            solblock = primal[block_num]
            if not surpluses and block_num == 0:
                total += block[0, 0, constraint] * solblock[0, 0]
                continue

            for row in xrange(block.shape[0]):
                for column in xrange(block.shape[1]):
                    val = block[row, column, constraint] * solblock[row, column]
                    total += val
        total -= constants[constraint - 1]
        v.append(total)

    return numpy.array(v, dtype=float)

def verify_solution(problem, primal, eps=1e-7, first_block = True, surpluses = True):
    eig = solution_eigenvalues(primal, first_block, surpluses)
    if any(eig < -eps):
        return False
    slack = solution_slackness(problem, primal, first_block, surpluses)
    slack = numpy.abs(slack)
    if any(slack > eps):
        return False
    return True

